/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	License:GPL
	---------------------------------------------------------------------------
	MDX sequencer(MDXDRV interface)
	seq_mdx.c
*/

#include "in_mdx.h"

/* ANSI/Windows standard headers */
#include <windows.h>
#include <shlobj.h>

/* Libraries headers */
#include "memmap_m.h"

/* Project headers */
#include "../sequencer.h"
#include "mdc2mdx.h"
#include "lzx042.h"
#include "spc.h"
#include "gym.h"

#define MAX_FILETYPES 32
#define MAX_TYP 128
#define MAX_EXT 128
#define MAX_OPT 256
#define MDXDRV_CMDLINE (MAX_PATH + MAX_OPT + MAX_PATH + 32)

enum MDC_STATE {
	MDC_OPEN,
	MDC_CLOSE
};

typedef struct {
	HANDLE hMap;
	volatile MdxdrvStatus *status;
	char cmdLine[MDXDRV_CMDLINE];
	LONG lock;
	unsigned int BootTime;
	enum MDC_STATE state;
	signed char id;
	int bufp;
} MDXDRVCONTROL;

void MDCInit(MDXDRVCONTROL *pmdc)
{
	pmdc->state = MDC_CLOSE;
	pmdc->hMap = NULL;
	pmdc->status = NULL;
	pmdc->id = '\xff';
	pmdc->lock = 1;
}

void MDCFreeControl(MDXDRVCONTROL *pmdc)
{
	if (pmdc->state != MDC_OPEN) return;
	if (pmdc->BootTime == pmdc->status->BootTime)
	{
		pmdc->status->ID[pmdc->id] = 0;
		InterlockedDecrement((LPLONG)&pmdc->status->InterLock);
	}
	UnmapViewOfFile((LPVOID)pmdc->status);
	CloseHandle(pmdc->hMap);
	pmdc->state = MDC_CLOSE;
	InterlockedExchange(&pmdc->lock, 1);	/* 䌠ԋp */
}

BOOL MDCGetControl(MDXDRVCONTROL *pmdc, BOOL bBoot)
{
	int time_out = -1;
	/* 䌠擾 */
	if (1 != InterlockedExchange(&pmdc->lock, 0)) return FALSE;
	do
	{
		if (time_out > 0)
		{
			Sleep(10);
			time_out -= 10;
		}
		pmdc->hMap = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, MapObjName);
		if (pmdc->hMap == NULL)
		{
			if (time_out < 0 && bBoot)
			{
				static STARTUPINFO si;
				static PROCESS_INFORMATION pi;
				char cwd[MAX_PATH];
				ZeroMemory(&si, sizeof(STARTUPINFO));
				ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
				si.cb = sizeof(STARTUPINFO);
				si.lpReserved = NULL;
				si.lpDesktop = NULL;
				si.lpTitle = NULL;
				si.dwFlags = STARTF_USESHOWWINDOW;
				si.wShowWindow = SW_HIDE;
				si.cbReserved2 = 0;
				si.lpReserved2 = NULL;
				GetCurrentDirectory(MAX_PATH, cwd);
				if (FALSE == CreateProcess(NULL, pmdc->cmdLine,
					NULL, NULL, FALSE,
					NORMAL_PRIORITY_CLASS,
					NULL,
					cwd,
					&si,
					&pi))
				{
					/* NɎsꍇ */
					InterlockedExchange(&pmdc->lock, 1);	/* 䌠ԋp */
					return FALSE;
				}
				/* Ȃ͎̂̂Ă */
				CloseHandle(pi.hThread);
				CloseHandle(pi.hProcess);
				/* Ƃ肠 500ms ҂ */
				Sleep(500);
				/*  500ms ҂ */
				time_out = 500;
			}
			continue;
		}
		pmdc->status = (MdxdrvStatus *)MapViewOfFile(pmdc->hMap, FILE_MAP_ALL_ACCESS, 0, 0, 0);
		if (pmdc->status == NULL)
		{
			CloseHandle(pmdc->hMap);
			/* NȂ҂ */
			continue;
		}
		time_out = 500;
		while (time_out > 0)
		{
			pmdc->BootTime = pmdc->status->BootTime;	/* NҔ */
			if (InterlockedIncrement((LPLONG)&pmdc->status->InterLock) == 0)
			{
				for (pmdc->id = 8; pmdc->id < memIDmax; pmdc->id++)
				{
					if (pmdc->status->ID[pmdc->id] != 0) continue;
					/* ID擾 */
					pmdc->status->ID[pmdc->id] = pmdc->id;
					/* 䌠ێ */
					pmdc->state = MDC_OPEN;
					return TRUE;
				}
				if (pmdc->BootTime == pmdc->status->BootTime)
					InterlockedDecrement((LPLONG)&pmdc->status->InterLock);
				UnmapViewOfFile((LPVOID)pmdc->status);
				CloseHandle(pmdc->hMap);
				InterlockedExchange(&pmdc->lock, 1);	/* 䌠ԋp */
				/* ID̋󂫂Ȃ */
				return FALSE;
			}
			if (pmdc->BootTime == pmdc->status->BootTime)
				InterlockedDecrement((LPLONG)&pmdc->status->InterLock);
			Sleep(10);
			time_out -= 10;
		}
		UnmapViewOfFile((LPVOID)pmdc->status);
		CloseHandle(pmdc->hMap);
		InterlockedExchange(&pmdc->lock, 1);	/* 䌠ԋp */
		/* ݌擾łȂ */
		return FALSE;
	} while (time_out > 0);
	InterlockedExchange(&pmdc->lock, 1);	/* 䌠ԋp */
	/* I[vłȂ */
	return FALSE;
}

BOOL MDCWait(MDXDRVCONTROL *pmdc, int load)
{
	int time_out = 500;
	if (pmdc->state != MDC_OPEN) return FALSE;
	while (pmdc->status->load != load)
	{
		if (time_out < 0) return FALSE;
		Sleep(10);
		time_out -= 10;
	}
	pmdc->bufp = 0;
	return TRUE;
}

BOOL MDCWaitInfinite(MDXDRVCONTROL *pmdc, int load)
{
	if (pmdc->state != MDC_OPEN) return FALSE;
	while (pmdc->status->load != load)
	{
		HANDLE hMutex;
		Sleep(10);
		/* I`FbN */
		hMutex = OpenMutex(MUTEX_ALL_ACCESS, FALSE, MutexObjName);
		if (hMutex == NULL) return FALSE;
		CloseHandle(hMutex);
	}
	pmdc->bufp = 0;
	return TRUE;
}

void MDCWriteDword(MDXDRVCONTROL *pmdc, DWORD dw)
{
	if (pmdc->state != MDC_OPEN) return;
	pmdc->status->buff[pmdc->bufp++] = (char)(dw >> 0x00);
	pmdc->status->buff[pmdc->bufp++] = (char)(dw >> 0x08);
	pmdc->status->buff[pmdc->bufp++] = (char)(dw >> 0x10);
	pmdc->status->buff[pmdc->bufp++] = (char)(dw >> 0x18);
}

BOOL MDCReadDword(MDXDRVCONTROL *pmdc, int *lpdw)
{
	if (pmdc->state != MDC_OPEN) return FALSE;
	*lpdw  = ((int)(unsigned char)pmdc->status->buff[pmdc->bufp++]) << 0x00;
	*lpdw |= ((int)(unsigned char)pmdc->status->buff[pmdc->bufp++]) << 0x08;
	*lpdw |= ((int)(unsigned char)pmdc->status->buff[pmdc->bufp++]) << 0x10;
	*lpdw |= ((int)(unsigned char)pmdc->status->buff[pmdc->bufp++]) << 0x18;
	return TRUE;
}

void MDCWriteString(MDXDRVCONTROL *pmdc, char *p)
{
	if (pmdc->state != MDC_OPEN) return;
	while (*p) pmdc->status->buff[pmdc->bufp++] = *(p++);
	pmdc->status->buff[pmdc->bufp++] = '\0';
}

/* MDX sequencer struct */
typedef struct {
	SEQUENCER seq;			/* KUMAamp common interface */
	MDXDRVCONTROL mdc;

	int isplaying;
	int volume;
	int panpot;
#if 0
	int loopc;				/* zero based loopcounter */
	int loopcount;			/* max loop count */
	int frequency;			/* mixing frequency */
	int fade;				/* 0:normal 1-:fading out(dB/(1 >> FADE_SHIFT)) */
	int fadetime;			/* fadeout time(ms) */
#endif
} MDXSEQ;
#define PTR__ ((MDXSEQ *)(this__->work))

/* amp setting struct */
static struct {
	char mdxdrvpath[MAX_PATH];
	struct {
		char typ[MAX_TYP];
		char ext[MAX_EXT];
		char opt[MAX_OPT];
	} filetypes[MAX_FILETYPES];

	int priority;

	HINSTANCE hDllInstance;
	char cfgname[MAX_PATH];
} setting;

static void GetPathNoExt(char *des, char *src)
{
	char *p = NULL;
	while (*src != '\0')
	{
		if (IsDBCSLeadByte(*src))
		{
			*(des++) = *(src++);
			*(des++) = *(src++);
			continue;
		}
		switch (*src)
		{
			case ':': case '\\': case '/':
				p = NULL;
				break;
			case '.':
				p = des;
				break;
		}
		*(des++) = *(src++);
	}
	if (p != NULL)
		*p = '\0';
	else
		*des = '\0';
}

static void GetPathExt(char *des, char *src)
{
	char *p = NULL;
	while (*src != '\0')
	{
		if (IsDBCSLeadByte(*src))
		{
			src += 2;
			continue;
		}
		switch (*src)
		{
			case ':': case '\\': case '/':
				p = NULL;
				break;
			case '.':
				p = src;
				break;
		}
		src++;
	}
	if (p != NULL)
		lstrcpy(des, p + 1);
	else
		*des = '\0';
}

static void Pause(SEQUENCER *this__, int ispause)
{
	if (!MDCGetControl(&PTR__->mdc, FALSE)) return;
	if (!MDCWait(&PTR__->mdc, 0))
	{
		MDCFreeControl(&PTR__->mdc);
		return;
	}
	MDCWriteDword(&PTR__->mdc, 1);	/* S̉t */
	MDCWriteDword(&PTR__->mdc, ispause ? 0 : 1);
	MDCWriteDword(&PTR__->mdc, 0);	/* functionI[ */
	PTR__->mdc.status->load = PTR__->mdc.id;
	if (MDCWaitInfinite(&PTR__->mdc, -PTR__->mdc.id))
	{
		PTR__->mdc.status->load = 0;
	}
	MDCFreeControl(&PTR__->mdc);
}

static void SetVolumePan(SEQUENCER *this__)
{
	int lvol, rvol;
	if (!MDCGetControl(&PTR__->mdc, FALSE)) return;
	if (!MDCWait(&PTR__->mdc, 0))
	{
		MDCFreeControl(&PTR__->mdc);
		return;
	}
	rvol = lvol = (100 * PTR__->volume) >> 7;
	if (PTR__->panpot > 0) lvol = (lvol * (127 - PTR__->panpot)) >> 7;
	if (PTR__->panpot < 0) rvol = (rvol * (127 + PTR__->panpot)) >> 7;
	MDCWriteDword(&PTR__->mdc, 11);	/* volume */
	MDCWriteDword(&PTR__->mdc, lvol);
	MDCWriteDword(&PTR__->mdc, rvol);
	MDCWriteDword(&PTR__->mdc, 0);	/* functionI[ */
	PTR__->mdc.status->load = PTR__->mdc.id;
	if (MDCWaitInfinite(&PTR__->mdc, -PTR__->mdc.id))
	{
		PTR__->mdc.status->load = 0;
	}
	MDCFreeControl(&PTR__->mdc);
}
static void SetVolume(SEQUENCER *this__, int volume)
{
	PTR__->volume = volume;
	SetVolumePan(this__);
}
static void SetPan(SEQUENCER *this__, int panpot)
{
	PTR__->panpot = panpot;
	SetVolumePan(this__);
}

static void Play(SEQUENCER *this__)
{
	if (!MDCGetControl(&PTR__->mdc, TRUE))
	{
		return;
	}
#if 0
	if (!MDCWait(&PTR__->mdc, 0))
	{
		MDCFreeControl(&PTR__->mdc);
		return;
	}
	MDCWriteDword(&PTR__->mdc, 4);	/* tԈʒuύX */
	MDCWriteDword(&PTR__->mdc, 1);	/* Ȑ擪 */
	MDCWriteDword(&PTR__->mdc, 0);	/* functionI[ */
	PTR__->mdc.status->load = PTR__->mdc.id;
	if (MDCWaitInfinite(&PTR__->mdc, -PTR__->mdc.id))
	{
		PTR__->mdc.status->load = 0;
	}
#endif
	MDCFreeControl(&PTR__->mdc);
	PTR__->isplaying = 1;
}

static void Stop(SEQUENCER *this__)
{
	PTR__->isplaying = 0;
	if (!MDCGetControl(&PTR__->mdc, FALSE)) return;
	if (!MDCWait(&PTR__->mdc, 0))
	{
		MDCFreeControl(&PTR__->mdc);
		return;
	}
	MDCWriteDword(&PTR__->mdc, 5);	/* I */
	MDCWriteString(&PTR__->mdc, MutexObjName);
	MDCWriteDword(&PTR__->mdc, 0);	/* functionI[ */
	PTR__->mdc.status->load = PTR__->mdc.id;
	if (MDCWaitInfinite(&PTR__->mdc, -PTR__->mdc.id))
	{
		PTR__->mdc.status->load = 0;
	}
	MDCFreeControl(&PTR__->mdc);
}

static void Term(SEQUENCER *this__)
{
	Stop(this__);
	free(PTR__);
}

static void SetPosition(SEQUENCER *this__, int time_in_ms)
{
	if (!MDCGetControl(&PTR__->mdc, FALSE)) return;
	if (!MDCWait(&PTR__->mdc, 0))
	{
		MDCFreeControl(&PTR__->mdc);
		return;
	}
	MDCWriteDword(&PTR__->mdc, 4);	/* tԈʒuύX */
	MDCWriteDword(&PTR__->mdc, 1 + time_in_ms);
	MDCWriteDword(&PTR__->mdc, 0);	/* functionI[ */
	PTR__->mdc.status->load = PTR__->mdc.id;
	if (MDCWaitInfinite(&PTR__->mdc, -PTR__->mdc.id))
	{
		PTR__->mdc.status->load = 0;
	}
	MDCFreeControl(&PTR__->mdc);
}
static int GetPosition(SEQUENCER *this__)
{
	int ret;
	if (!PTR__->isplaying) return 0;
	if (!MDCGetControl(&PTR__->mdc, FALSE)) return 0;
	if (!MDCWait(&PTR__->mdc, 0))
	{
		MDCFreeControl(&PTR__->mdc);
		return 0;
	}
	MDCWriteDword(&PTR__->mdc, 4);	/* tԈʒuύX */
	MDCWriteDword(&PTR__->mdc, 0);	/* tԈʒu߂ */
	MDCWriteDword(&PTR__->mdc, 0);	/* functionI[ */
	PTR__->mdc.status->load = PTR__->mdc.id;
	if (!MDCWaitInfinite(&PTR__->mdc, -PTR__->mdc.id))
	{
		MDCFreeControl(&PTR__->mdc);
		return 0;
	}
	if (
		!MDCReadDword(&PTR__->mdc, &ret) ||
		ret != 4 ||
		!MDCReadDword(&PTR__->mdc, &ret) ||
		ret < 8 ||
		!MDCReadDword(&PTR__->mdc, &ret) ||
		ret != 0 ||
		!MDCReadDword(&PTR__->mdc, &ret)
	)
		ret = 0;
	PTR__->mdc.status->load = 0;
	MDCFreeControl(&PTR__->mdc);
	return ret;
}

static int GetLoopCount(SEQUENCER *this__)
{
	return 0;
}

static int IsPlaying(SEQUENCER *this__)
{
	int ret;
	if (!PTR__->isplaying) return 0;
	if (!MDCGetControl(&PTR__->mdc, FALSE)) return 0;
	if (!MDCWait(&PTR__->mdc, 0))
	{
		MDCFreeControl(&PTR__->mdc);
		return 0;
	}
	MDCWriteDword(&PTR__->mdc, 1);	/* S̉t */
	MDCWriteDword(&PTR__->mdc, -1);	/* Ԃ𒲂ׂ */
	MDCWriteDword(&PTR__->mdc, 0);	/* functionI[ */
	PTR__->mdc.status->load = PTR__->mdc.id;
	if (!MDCWaitInfinite(&PTR__->mdc, -PTR__->mdc.id))
	{
		MDCFreeControl(&PTR__->mdc);
		return 0;
	}
	if (
		!MDCReadDword(&PTR__->mdc, &ret) ||
		ret != 1 ||
		!MDCReadDword(&PTR__->mdc, &ret) ||
		ret < 8 ||
		!MDCReadDword(&PTR__->mdc, &ret) ||
		ret != 0 ||
		!MDCReadDword(&PTR__->mdc, &ret)
	)
		ret = 0;
	else
		ret = 1;
	PTR__->mdc.status->load = 0;
	MDCFreeControl(&PTR__->mdc);
	return ret;
}

static int GetPriority(SEQUENCER *this__)
{
	return setting.priority;
}

static int MatchExt(char *ext, char *exts)
{
	char *p;
	char buf[MAX_EXT];
	while (*exts != '\0')
	{
		p = buf;
		while (*exts != ';' && *exts != '\0')
		{
			*p++ = *exts++;
		}
		*p = '\0';
		if (*exts == ';') exts++;
		if (lstrcmpi(ext, buf) == 0) return 1;
	}
	return 0;
}

static int MakeMdxdrvComline(char *cmdLine, char *fn)
{
	int i;
	char ext[MAX_EXT];
	if (*fn == '\0') return 0;
	lstrcpy(cmdLine, setting.mdxdrvpath);
	GetPathExt(ext, fn);
	for (i = 0; i < MAX_FILETYPES; i++)
	{
		if (MatchExt(ext, setting.filetypes[i].ext))
		{
			if (setting.filetypes[i].opt[0] != '\0')
			{
				lstrcat(cmdLine," ");
				lstrcat(cmdLine,setting.filetypes[i].opt);
			}
			lstrcat(cmdLine," \"");
			lstrcat(cmdLine,fn);
			lstrcat(cmdLine,"\"");
			return 1;
		}
	}
	return 0;
}

#ifdef SINGLE_BUILD
#define InitMdxSequencer InitSequencer
#define QuitMdxSequencer QuitSequencer
#define AboutMdxSequencer AboutSequencer
#define ConfigMdxSequencer ConfigSequencer
#define getMdxFileInfo getFileInfo
#define loadMdxFile loadFile
#define infoMdxBox infoBox
#define isOurMdxFile isOurFile
#endif

SEQUENCER *loadMdxFile(char *fn_)
{
	char ext[MAX_EXT];
	char *fn;
	MDXSEQ *p;

	mdcclean();
	lzxclean();
	gymclean();

	p = (MDXSEQ *)malloc(sizeof(MDXSEQ));
	if (p == NULL) return NULL;

	GetPathExt(ext, fn_);
	if (MatchExt(ext, "MDX"))
	{
		fn = lzxdecode(fn_);
	}
	else if (MatchExt(ext, "MDC"))
	{
		fn = mdc2mdx(fn_);
	}
	else if (MatchExt(ext, "GYM"))
	{
		fn = gyminflate(fn_);
	}
	else
	{
		fn = fn_;
	}

	if (!MakeMdxdrvComline(p->mdc.cmdLine, fn))
	{
		free(p);
		return NULL;
	}
	MDCInit(&p->mdc);
	p->isplaying = 0;
	p->volume = 128;
	p->panpot = 0;
	p->seq.work = p;
	p->seq.Play = Play;
	p->seq.Stop = Stop;
	p->seq.Pause = Pause;
	p->seq.SetVolume = SetVolume;
	p->seq.SetPan = SetPan;
	p->seq.Term = Term;
	p->seq.SetPosition = SetPosition;
	p->seq.GetPosition = GetPosition;
	p->seq.GetPriority = GetPriority;
	p->seq.GetLoopCount = GetLoopCount;
	p->seq.IsPlaying = IsPlaying;
	return &p->seq;
}

int infoMdxBox(char *fn, HWND hwndParent)
{
	return 0;
}
int isOurMdxFile(char *fn)
{
	return 0;
}

void InitMdxSequencer(HINSTANCE hDllInstance)
{
	setting.priority = 3;
}
void QuitMdxSequencer(HINSTANCE hDllInstance)
{
	mdcclean();
	lzxclean();
	gymclean();
}
void AboutMdxSequencer(HWND hwndParent)
{
	MessageBox(hwndParent, "amp\n\t(C) 2000 Mamiya\n\t\tmailto:mamiya@proc.org.tohoku.ac.jp\n\t\thttp://www.proc.org.tohoku.ac.jp/befis/", PLUGIN_NAME, MB_OK);
}
void ConfigMdxSequencer(HWND hwndParent)
{
}

int getMdxFileInfo(char *fn, char *title, int *length_in_ms)
{
	char ext[MAX_EXT];
	if (title) *title = '\0';
	if (length_in_ms) *length_in_ms = 0;
	GetPathExt(ext, fn);
	if (MatchExt(ext, "MDX"))
	{
		if (title) GetTitleMDX(fn, title);
		/* if (length_in_ms) GetTimeMDX(fn, length_in_ms); */
	}
	else if (MatchExt(ext, "MDC"))
	{
		if (title) GetTitleMDC(fn, title);
		/* if (length_in_ms) GetTimeMDX(fn, length_in_ms); */
	}
	else if (MatchExt(ext, "SPC"))
	{
		if (title) GetTitleSPC(fn, title);
		if (length_in_ms) GetTimeSPC(fn, length_in_ms);
	}
	else if (MatchExt(ext, "GYM"))
	{
		if (title) GetTitleGYM(fn, title);
#if 0
		if (length_in_ms) GetTimeGYM(fn, length_in_ms);
#endif
	}
	if (title && *title == '\0')
	{
		lstrcpy(title, "(");
		lstrcat(title, fn);
		lstrcat(title, ")");
	}
	if (length_in_ms && *length_in_ms == 0)
	{
		*length_in_ms = 5 * 60 * 1000;	/* Ƃ肠5 */
	}
	return 0;
}

static void GetSettingString(LPCSTR lpKey, LPCSTR lpDefault, LPSTR lpBuf, DWORD nSize)
{
	GetPrivateProfileString("MDXDRV", lpKey, lpDefault, lpBuf, nSize, setting.cfgname);
}
static void SetSettingString(LPCSTR lpKey, LPCSTR lpStr)
{
	WritePrivateProfileString("MDXDRV", lpKey, lpStr, setting.cfgname);
}


BOOL WINAPI DllMain(
	HINSTANCE hinstDLL,
	DWORD fdwReason,
	LPVOID lpvReserved
)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		int i, numtypes = 0;
		char *p = FileTypes;
		char cfgbuf[MAX_PATH];
		LPSTR fnoff;
		setting.hDllInstance = hinstDLL;
		GetModuleFileName(hinstDLL, setting.cfgname, MAX_PATH);
		GetFullPathName(setting.cfgname, MAX_PATH, cfgbuf, &fnoff);
		GetPathNoExt(setting.cfgname, cfgbuf);
		lstrcat(setting.cfgname, ".ini");
		GetSettingString("MDXDRVPath","C:\\Program Files\\Winamp\\Plugins\\mdxdrv.exe /A2",setting.mdxdrvpath, MAX_PATH);
		for (i = 0; i < MAX_FILETYPES; i++)
		{
			char key[32];
			wsprintf(key, "TYP_%d", i);
			GetSettingString(key,"",setting.filetypes[numtypes].typ, MAX_TYP);
			if (setting.filetypes[numtypes].typ[0] == '\0') continue;
			wsprintf(key, "EXT_%d", i);
			GetSettingString(key,"",setting.filetypes[numtypes].ext, MAX_EXT);
			wsprintf(key, "OPT_%d", i);
			GetSettingString(key,"",setting.filetypes[numtypes].opt, MAX_OPT);
			lstrcpy(p, setting.filetypes[numtypes].ext);
			p += lstrlen(setting.filetypes[numtypes].ext) + 1;
			lstrcpy(p, setting.filetypes[numtypes].typ);
			p += lstrlen(setting.filetypes[numtypes].typ) + 1;
			numtypes++;
		}
		if (numtypes == 0)
		{
			/* default */
			lstrcpy(setting.filetypes[numtypes].typ, "MXDRV16y music file(*.mdx)");
			lstrcpy(setting.filetypes[numtypes].ext, "MDX");
			lstrcpy(setting.filetypes[numtypes].opt, "/X");
			lstrcpy(p, setting.filetypes[numtypes].ext);
			p += lstrlen(setting.filetypes[numtypes].ext) + 1;
			lstrcpy(p, setting.filetypes[numtypes].typ);
			p += lstrlen(setting.filetypes[numtypes].typ) + 1;
			numtypes++;
		}
		*p = '\0';
	}
	return TRUE;
}
