/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	---------------------------------------------------------------------------
	generic WINAMP interface
	---------------------------------------------------------------------------
	$Id: in_xxx.h,v 0.7.6 2000/03/27 00:32:53 Mamiya Exp $
*/

#define MAX_SAMPLERATE 48000

static In_Module mod;
static volatile struct {
	HANDLE hEvent;
	HANDLE hMutex;
	int request;
	int paused;
	int need_seek;
	int length;
	int time_pos;
	int need_panpot;
	int need_volume;
	int isstoped;
	char fn[MAX_PATH];
} player_data = {
	NULL,NULL,
	0,0,0,0,0,
	0,128,
	1,
	""
};

enum PLAYER_REQUEST {
	PLAYER_REQUEST_QUIT,
	PLAYER_REQUEST_STOP,
	PLAYER_REQUEST_PANPOT,
	PLAYER_REQUEST_VOLUME,
	PLAYER_REQUEST_PAUSE,
	PLAYER_REQUEST_UNPAUSE,
	PLAYER_REQUEST_SEEK
};

static void PlayerSetEvent(enum PLAYER_REQUEST request)
{
	if (player_data.hEvent != NULL)
	{
		player_data.request |= 1 << request;
		SetEvent(player_data.hEvent);
	}
}

static void init(void)
{
	player_data.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	player_data.hMutex = CreateEvent(NULL, TRUE, TRUE, NULL);
	InitSequencer(mod.hDllInstance);
}

static void quit(void)
{
	QuitSequencer(mod.hDllInstance);
	if (player_data.hEvent != NULL) {
		CloseHandle(player_data.hEvent);
		player_data.hEvent = NULL;
	}
	if (player_data.hMutex != NULL) {
		CloseHandle(player_data.hMutex);
		player_data.hMutex = NULL;
	}
}

static void getfileinfo(char *fn_, char *title, int *length_in_ms)
{
	static LONG lock = 1;
	while (1 != InterlockedExchange(&lock, 0)) Sleep(10);
	if (fn_ == NULL || *fn_ == '\0') fn_ = (char *)player_data.fn;
	getFileInfo(fn_, title, length_in_ms);
	InterlockedExchange(&lock, 1);
}

static unsigned long PlayThread(SEQUENCER *sequencer)
{
#if !N_VERSION
#if !defined(MAX_SAMPLERATE)
	void *samp_buf = NULL;	/* DYNAMIC ALLOCATION */
#else
	static short samp_buf[((MAX_SAMPLERATE * 30) / 1000) * 2 * 2 + 8];
#endif
#endif
	int request = 0;
	player_data.paused = 0;
	player_data.request = 0;
	ResetEvent(player_data.hEvent);
	/* Đ͏ɔVOi */
	ResetEvent(player_data.hMutex);
	/* try */
	do
	{
		static int tbl_prio[] = {
			THREAD_PRIORITY_IDLE,
			THREAD_PRIORITY_LOWEST,
			THREAD_PRIORITY_BELOW_NORMAL,
			THREAD_PRIORITY_NORMAL,
			THREAD_PRIORITY_ABOVE_NORMAL,
			THREAD_PRIORITY_HIGHEST,
			THREAD_PRIORITY_TIME_CRITICAL
		};
		int samplerate, loop_n, loop_c = -1;
		DWORD wt;
#if !N_VERSION
		int bufsize, len;
		samplerate = sequencer->GetRate(sequencer);
		bufsize = (samplerate * 30) / 1000;	/* 30ms */
#if !defined(MAX_SAMPLERATE)
		samp_buf = malloc(bufsize * 4 * 2 + 8);
		if (samp_buf == NULL) break;
#endif
#else
		samplerate = 0;
#endif
		SetThreadPriority(GetCurrentThread(), tbl_prio[sequencer->GetPriority(sequencer)]);
		sequencer->Play(sequencer);

		while (sequencer->IsPlaying(sequencer))
		{
			player_data.time_pos = sequencer->GetPosition(sequencer);

			loop_n = sequencer->GetLoopCount(sequencer);
			if (loop_c != loop_n)
			{
				loop_c = loop_n;
				mod.SetInfo(loop_c, samplerate / 1000, 2, 0);
			}

#if !N_VERSION
			sequencer->Mix(sequencer, samp_buf, bufsize);
			mod.SAAddPCMData(samp_buf, 2, 16, player_data.time_pos);
			mod.VSAAddPCMData(samp_buf, 2, 16, player_data.time_pos);
			if(mod.dsp_isactive())
			{
				len = mod.dsp_dosamples(samp_buf, bufsize, 16, 2, samplerate);
			}
			else
			{
				len = bufsize;
			}
			wt = 0;
#else
			wt = 10;
#endif
			do
			{
				if (WaitForSingleObject(player_data.hEvent, wt) != WAIT_TIMEOUT)
				{
					request = InterlockedExchange((LPLONG)&player_data.request, 0);
					if (request == 0)
					{
						/* ??? THREAD FAULT ??? */
						/* Cȃ^C~OŃ{^ */
						request |= (1 << PLAYER_REQUEST_STOP);
					}
					break;
				}
				wt = 10;
			}
#if !N_VERSION
			while (mod.outMod->CanWrite() < (len << 2));
#else
			while (0);
#endif
			if (request & (1 << PLAYER_REQUEST_QUIT)) break;
			if (request & (1 << PLAYER_REQUEST_STOP)) break;
#if N_VERSION
			if (request & (1 << PLAYER_REQUEST_PANPOT))
			{
				sequencer->SetPan(sequencer, player_data.need_panpot);
				request -= (1 << PLAYER_REQUEST_PANPOT);
			}
			if (request & (1 << PLAYER_REQUEST_VOLUME))
			{
				sequencer->SetVolume(sequencer, player_data.need_volume);
				request -= (1 << PLAYER_REQUEST_VOLUME);
			}
			if (request & (1 << PLAYER_REQUEST_PAUSE))
			{
				player_data.paused = 1;
				sequencer->Pause(sequencer, player_data.paused);
				request = 0;
				continue;
			}
			if (request & (1 << PLAYER_REQUEST_UNPAUSE))
			{
				player_data.paused = 0;
				sequencer->Pause(sequencer, player_data.paused);
				request = 0;
				continue;
			}
#endif
			if (request & (1 << PLAYER_REQUEST_SEEK))
			{
#if !N_VERSION
				mod.outMod->Flush(player_data.need_seek);
#endif
				sequencer->SetPosition(sequencer, player_data.need_seek);
				player_data.need_seek = 0;
				request = 0;
				continue;
			}
#if !N_VERSION
			if (mod.outMod->CanWrite() >= (len << 2))
				mod.outMod->Write((char *)samp_buf, (len << 2));
#endif
		}
#if !N_VERSION
		if (request == 0)
		{
			/*  ĐI(o͏I҂) */
			wt = 0;
			do
			{
				if (WaitForSingleObject(player_data.hEvent, wt) != WAIT_TIMEOUT)
				{
					request = InterlockedExchange((LPLONG)&player_data.request, 0);
					if (request & (1 << PLAYER_REQUEST_QUIT)) break;
					if (request & (1 << PLAYER_REQUEST_STOP)) break;
					/* ??? THREAD FAULT ??? */
					/* Cȃ^C~OŃ{^ */
					request |= (1 << PLAYER_REQUEST_STOP);
					break;
				}
				wt = 10;
			} while (mod.outMod->IsPlaying());
		}
#endif
		sequencer->Stop(sequencer);
	} while (0);
	/* finaly */

	sequencer->Term(sequencer);
#if !N_VERSION
#if !defined(MAX_SAMPLERATE)
	if (samp_buf != NULL) free(samp_buf);
#endif
#endif

	if (!request)
	{
		/* WINAMPɏIʒm */
#define WM_WA_MPEG_EOF (WM_USER + 2)
		PostMessage(mod.hMainWindow, WM_WA_MPEG_EOF, 0, 0);
	}

	InterlockedExchange((LPLONG)&player_data.isstoped, 1);
	/* ҋ@XbhɏIʒm */
	SetEvent(player_data.hMutex);
	return 0;
}

static int play(char *fn)
{
	if (1 != InterlockedExchange((LPLONG)&player_data.isstoped, 0))
		return 1;
	do
	{
		SEQUENCER *sequencer;
#if !N_VERSION
		int samplerate, maxlatency;
#endif
		lstrcpy((char *)player_data.fn, fn);
		getfileinfo(NULL, NULL, (int *)&player_data.length);
		sequencer = loadFile(fn);
#if !N_VERSION
		samplerate = (sequencer != NULL)?sequencer->GetRate(sequencer):44100;
		maxlatency = mod.outMod->Open(samplerate, 2, 16, 0, 0);
		if (maxlatency < 0)
		{
			if (sequencer != NULL) sequencer->Term(sequencer);
			break;
		}
		mod.SetInfo(0, samplerate / 1000, 2, 0);
		mod.SAVSAInit(samplerate, maxlatency);
		mod.VSASetInfo(2, samplerate);
		mod.outMod->SetVolume(-666);
#endif

		if (player_data.hEvent == NULL || player_data.hMutex == NULL || sequencer == NULL)
		{
			if (sequencer != NULL) sequencer->Term(sequencer);
			InterlockedExchange((LPLONG)&player_data.isstoped, 1);
		}
		else
		{
			/* ĐXbh̐ */
			_beginthread(PlayThread, 0, sequencer);
		}
		return 0;
	} while(0);
	InterlockedExchange((LPLONG)&player_data.isstoped, 1);
	return 1;
}

static void stop(void)
{
	static LONG lock = 1;
	while (1 != InterlockedExchange(&lock, 0)) return;
	while (!player_data.isstoped)
	{
		PlayerSetEvent(PLAYER_REQUEST_STOP);
		WaitForSingleObject(player_data.hMutex, 10);
	}
#if !N_VERSION
	mod.outMod->Close();
	mod.SAVSADeInit();
#endif
	InterlockedExchange(&lock, 1);
}

#if !N_VERSION
static void pause(void) { player_data.paused = 1; mod.outMod->Pause(1); }
static void unpause(void) { player_data.paused = 0; mod.outMod->Pause(0); }
static int getoutputtime(void) { return mod.outMod->GetOutputTime(); }
static void setpan(int pan) { mod.outMod->SetPan(pan); }
static void setvolume(int volume) { mod.outMod->SetVolume(volume); }
#else
static void pause(void) { PlayerSetEvent(PLAYER_REQUEST_PAUSE); }
static void unpause(void) { PlayerSetEvent(PLAYER_REQUEST_UNPAUSE); }
static int getoutputtime(void) { return player_data.time_pos; }
static void setpan(int pan)
{
	player_data.need_panpot = pan;
	PlayerSetEvent(PLAYER_REQUEST_PANPOT);
}
static void setvolume(int volume)
{
	player_data.need_volume = volume;
	PlayerSetEvent(PLAYER_REQUEST_VOLUME);
}
#endif

static void config(HWND hwnd) { ConfigSequencer(hwnd); }
static void about(HWND hwnd) { AboutSequencer(hwnd); }
static int infobox(char *fn, HWND hwnd) { return infoBox(fn, hwnd); }
static int isourfile(char *fn) { return isOurFile(fn); }
static void eq_set(int on, char data[10], int preamp) {}
static int ispaused(void) { return player_data.paused; }
static int getlength(void) { return player_data.length; }
static void setoutputtime(int time_in_ms)
{
	player_data.need_seek = time_in_ms;
	PlayerSetEvent(PLAYER_REQUEST_SEEK);
}

static In_Module mod = {
	IN_VER,
	PLUGIN_NAME,
	0,	/* hMainWindow */
	0,	/* hDllInstance */
	PLUGIN_TYPES,
	1,	/* is_seekable */
#if !N_VERSION
	1,	/* UsesOutputPlug */
#else
	0,
#endif
	config,
	about,
	init,
	quit,
	getfileinfo,
	infobox,
	isourfile,
	play,
	pause,
	unpause,
	ispaused,
	stop,

	getlength,
	getoutputtime,
	setoutputtime,
	setvolume,
	setpan,

	/* pointers filled in by winamp */
	0,0,0,0,0,0,0,0,0,	/* vis stuff */
	0,0,	/* dsp */

	eq_set,	/* EQ, not used */
	NULL,	/* setinfo */
	0		/* out_mod */
};

In_Module __declspec(dllexport) *winampGetInModule2(void)
{
	return &mod;
}

